import datetime

DATE_FORMAT = "%Y-%m-%d"
TIME_FORMAT = "%Y-%m-%d %H:%M"
TIME_FORMAT_SECONDS = "%Y-%m-%d %H:%M:%S"
TIME_FORMAT_MILLISECONDS = "%Y-%m-%d %H:%M:%S:%f"


def get_today_date_string():
    current_date = datetime.datetime.strftime(datetime.datetime.now(), DATE_FORMAT)

    return current_date


def get_today_time_string():
    current_date = datetime.datetime.strftime(datetime.datetime.now(), TIME_FORMAT)

    return current_date


def get_today_time_seconds_string():
    current_date = datetime.datetime.strftime(datetime.datetime.now(), TIME_FORMAT_SECONDS)

    return current_date


def get_today_time_milliseconds_string():
    current_date = datetime.datetime.strftime(datetime.datetime.now(), TIME_FORMAT_MILLISECONDS)

    return current_date




def select(options, title=""):
    if len(options) == 1:
        return 0

    import simple_term_menu

    return simple_term_menu.TerminalMenu(options, title=title).show()


def select_yn(title="") -> bool:
    selected_option_id = select(["[n]No", "[y]Yes"], title)

    return bool(selected_option_id)  # bool hack: 0 and 1 match False and True


def get_input_using_temporary_file(file_path=".tmp", editor="vim", initial_message="", force_rewrite_initial_message=True):
    import tired.command
    import os

    # Create file
    if not os.path.isfile(file_path) or force_rewrite_initial_message:
        with open(file_path, 'w') as f:
            f.write(initial_message)

    tired.command.execute(f"{editor} {file_path}")

    with open(file_path, 'r') as f:
        return f.read()



"""
Frequently reused Tkinter snippets for building one-off GUI
applications in the most efficient way
"""

# TODO: add interface for easily building simple GUIs with checkboxes, tabs, file selection dialogs, drop-down selections, menus, and buttons

from tkinter import ttk
import threading
import tired.logging
import tkinter


DEFAULT_LABEL_WIDTH = 25


class FileSelectionWidget(tkinter.Frame):

    def __init__(self, parent, title, path_string_variable: tkinter.StringVar, mode_open=True, file_types=("*.*",), *args, **kwargs):
        from tkinter.filedialog import askopenfilename, asksaveasfilename
        # TODO: save value variable in "self"

        tkinter.Frame.__init__(self, parent, *args, **kwargs)

        file_types = tuple([(i,i,) for i in file_types])
        file_dialog_cb = askopenfilename if mode_open else asksaveasfilename

        self.mode_open = mode_open
        self.title_label = tkinter.Label(self, text=title, width=DEFAULT_LABEL_WIDTH, anchor='w')
        self.path_label = tkinter.Label(self, text="", textvariable=path_string_variable)
        self.select_path_btn = tkinter.Button(self, text="...", command=lambda: path_string_variable.set(file_dialog_cb(title=title, filetypes=file_types)))
        self.reset_btn = tkinter.Button(self, text="Reset", command=lambda: path_string_variable.set(""))

        self.title_label.grid(row=0, column=0, sticky='w')
        self.select_path_btn.grid(row=0, column=1)
        self.reset_btn.grid(row=0, column=2)
        self.path_label.grid(row=0, column=3, sticky='wn')


class LabeledSpinbox(tkinter.Frame):
    def __init__(self, parent, title, textvariable, from_=1.0, to=10.0, increment=1.0, *args, **kwargs):
        # TODO: save value variable in "self"
        tkinter.Frame.__init__(self, parent)
        self.label = tkinter.Label(self, text=title, width=DEFAULT_LABEL_WIDTH, anchor='w')
        self.spinbox = tkinter.Spinbox(self, *args, textvariable=textvariable, from_=from_, to=to, increment=increment, **kwargs)
        self.label.grid(row=0, column=0, sticky='nsew')
        self.spinbox.grid(row=0, column=1, sticky='nsew')


class LabeledOptionsMenu(tkinter.Frame):
    def __init__(self, parent, title, targetvar, *args, **kwargs):
        # TODO: save value variable in "self"
        tkinter.Frame.__init__(self, parent)

        self.label = tkinter.Label(self, text=title, width=DEFAULT_LABEL_WIDTH, anchor='w')
        self.option_menu = tkinter.OptionMenu(self, targetvar, *args, **kwargs)
        self.label.grid(row=0, column=0, sticky='nsew')
        self.option_menu.grid(row=0, column=1, sticky='nsew')


class GridPlacementStrategy:
    def place_widget(self, parent: tkinter.Widget, widget: tkinter.Widget):
        nrow = len(parent.grid_slaves())
        tired.logging.debug(f"Adding widget {widget.__class__.__name__} at row {nrow}")
        widget.grid(row=nrow, column=0, sticky='w')


class Frame(tkinter.Frame):
    def __init__(self, parent, *args, **kwargs):
        tkinter.Frame.__init__(self, parent, *args, **kwargs)
        self._tk_variables_map = dict()
        self._file_dialog_map = dict()
        self._checkbox_map = dict()
        self._button_map = dict()
        self._spinbox_map = dict()
        self._tabs_map = dict()
        self._frame_map = dict()

        # Defines how widgets should be arranged on a pane
        self._placement_strategy = GridPlacementStrategy()

    def _is_widget_registered(self, widget_name):
        return widget_name in self._tk_variables_map.keys() or widget_name in self._tabs_map.keys() or widget_name in self._frame_map.keys()

    def add_checkbox(self, string_identifier: str, default_value=False):
        """
        Adds a checkbox onto pane. String identifier is used as title
        """
        if self._is_widget_registered(string_identifier):
            raise KeyError(f"A widget with the name \"{string_identifier}\" already exists")

        variable = tkinter.BooleanVar()
        self._tk_variables_map[string_identifier] = variable
        widget = tkinter.Checkbutton(self, text=string_identifier, variable=variable)
        self._checkbox_map[string_identifier] = widget
        self._placement_strategy.place_widget(self, widget)

        return widget

    def set_widget_value(self, widget_string_identifier: str, value: object):
        """
        Sets value to a variable corresponding to a widget selected
        """
        if not self._is_widget_registered(widget_string_identifier):
            raise KeyError(f"A widget with the name \"{widget_string_identifier}\" does not exist")

        self._tk_variables_map[widget_string_identifier].set(value)

    def add_file_selection(self, string_identifier: str, file_types=('*.*', )):
        """
        Adds file dialog onto plane.
        """
        if self._is_widget_registered(string_identifier):
            raise KeyError(f"A widget with the name \"{string_identifier}\" already exists")

        variable = tkinter.StringVar()
        self._tk_variables_map[string_identifier] = variable
        widget = FileSelectionWidget(self, string_identifier, variable, True, file_types)
        self._file_dialog_map[string_identifier] = widget
        self._placement_strategy.place_widget(self, widget)

        return widget

    def add_spinbox(self, string_identifier: str, min: float, max: float, step: float = 1.0):
        """
        Adds spinbox onto plane.
        post: The value type for spinbox is `double`
        """
        if self._is_widget_registered(string_identifier):
            raise KeyError(f"A widget with the name \"{string_identifier}\" already exists")

        variable = tkinter.DoubleVar()
        self._tk_variables_map[string_identifier] = variable
        widget = LabeledSpinbox(self, string_identifier, variable, min, max, step)
        self._spinbox_map[string_identifier] = widget
        self._placement_strategy.place_widget(self, widget)

        return widget

    def add_button(self, string_identifier, callback=lambda: None):
        """
        Adds a simple button
        """
        if self._is_widget_registered(string_identifier):
            raise KeyError(f"A widget with the name \"{string_identifier}\" already exists")

        widget = tkinter.Button(self, text=string_identifier, command=callback)
        self._button_map[string_identifier] = widget
        self._placement_strategy.place_widget(self, widget)

        return widget

    def add_tabs(self, string_identifier):
        from tired.tk import Tabs

        if self._is_widget_registered(string_identifier):
            raise KeyError(f"A widget with the name \"{string_identifier}\" already exists")

        widget = Tabs(self)
        self._placement_strategy.place_widget(self, widget)
        self._tabs_map[string_identifier] = widget

        return widget

    def add_frame(self, string_identifier):
        if self._is_widget_registered(string_identifier):
            raise KeyError(f"A widget with the name \"{string_identifier}\" already exists")

        widget = Frame(self)
        self._frame_map[string_identifier] = widget
        self._placement_strategy.place_widget(self, widget)

        return widget


class Tabs(ttk.Notebook):
    def __init__(self, *args, **kwargs):
        ttk.Notebook.__init__(self, *args, **kwargs)
        self._frame_map = dict()

    def _is_widget_registered(self, widget_string_identifier: str):
        return widget_string_identifier in self._frame_map.keys()

    def add_frame(self, string_identifier):
        if self._is_widget_registered(string_identifier):
            raise KeyError(f"A widget with the name \"{string_identifier}\" already exists")

        widget = Frame(self)
        self.add(widget, text=string_identifier)

        return widget


class Window(tkinter.Tk):
    def __init__(self, *args, **kwargs):
        tkinter.Tk.__init__(self, *args, **kwargs)
        self._root_frame = Frame(self)
        self._root_frame.pack(expand=True, fill='both')

    def _is_widget_registered(self, widget_string_identifier: str):
        return widget_string_identifier in self._frame_map.keys() or widget_string_identifier in self._tabs_map.keys()

    def add_frame(self, string_identifier: str):
        return self._root_frame.add_frame(string_identifier)

    def add_tabs(self, string_identifier: str):
        return self._root_frame.add_tabs(string_identifier)

    def run(self):
        self.mainloop()



import tired.command
import tired.logging
import tired.parse


_LOG_CONTEXT = "tired.command"


def get_current_branch_name():
    command_string = "git rev-parse --abbrev-ref HEAD"
    output, code = tired.command.get_output_with_code(command_string)

    if code != 0:
        tired.logging.error(_LOG_CONTEXT, f"Failed to execute command `{command_string}`")

        raise Exception(f"Failed to execute command `{command_string}`, returned {code}")

    return output.strip()


def get_current_commit_hash():
    command_string = "git rev-parse HEAD"
    output, code = tired.command.get_output_with_code(command_string)

    if code != 0:
        tired.logging.error(_LOG_CONTEXT, f"Failed to execute command `{command_string}`")

        raise Exception(f"Failed to execute command `{command_string}`, returned {code}")

    return output.strip()


def get_staged_file_paths(use_relative_paths=False):
    """
    @param use_relative_paths. If true, the returned paths will be relative to PWD
    """
    relative_flag = "--relative" if use_relative_paths else ""
    command_string = f"git diff --name-only --staged {relative_flag}"
    output, code = tired.command.get_output_with_code(command_string)

    if code != 0:
        tired.logging.error(_LOG_CONTEXT, f"Failed to execute command `{command_string}`")

        raise Exception(f"Failed to execute command `{command_string}`, returned {code}")

    output = output.strip()

    return tired.parse.iterate_string_multiline(output)


def get_git_directory_from_nested_context():
    """
    TODO: git rev-parse --show-toplevel
    """
    command_string = f"git rev-parse --show-toplevel"
    output, code = tired.command.get_output_with_code(command_string)

    if code != 0:
        tired.logging.error(f"Failed to execute command `{command_string}`")

        raise Exception(f"Failed to execute command `{command_string}`, returned {code}")

    output = output.strip()

    return output



import inspect


def module_file_as_module_object(module_file):
    """
    Example:

    ```
    module = tired.meta.module_file_as_module_object(__file__)
    ```
    """

    #TODO
    pass


def get_module_functions(module_object):
    # TODO
    pass


def get_stack_context_string(caller_stack_level=1):
    """
    Builds the caller's context using instrospection based on "inspect". The
    format is this:

    <MODULE>.<CLASS_NAME_INCLUDING_NESTED_ONES>.<FUNCTION_NAME>

    README: Implementation guidelines

    Get nested class name

    ```
    import inspect

    def get_caller_class_name():
        stack = inspect.stack()
        caller_frame = stack[1]
        caller_class_name = caller_frame[0].f_locals.get("self", None).__class__.__name__
        return caller_class_name
    ```

    Getting function name from that is obvious

    """
    # TODO
    stack = inspect.stack()
    caller_frame = stack[caller_stack_level]

    try:
        qual_name = caller_frame[0].f_code.co_qualname
    except AttributeError:
        qual_name = caller_frame[0].f_code.co_name  # TODO handle call from class instance (use `self` variable)

    module_name = inspect.getmodule(caller_frame[0]).__name__

    return f"{module_name}.{qual_name}"



def get_directory_content(directory: str):
    """
    Iterate only through directory content
    """
    import os
    import pathlib

    real_path = pathlib.Path(directory).resolve()
    content = os.listdir(str(real_path))

    return content


def get_directory_content_directories(directory: str, exclude_symbolic_links=False):
    """
    List-out everything but directories
    """
    import os
    import pathlib

    directory_path_object = pathlib.Path(directory).resolve()

    for item in get_directory_content(directory):
        absolute_path_string = str(directory_path_object / item)

        if os.path.isdir(absolute_path_string) \
                and not (exclude_symbolic_links and os.path.islink(absolute_path_string)):
            yield item




import serial


class SerialStream:
    def __init__(self, dev, rate):
        self.socket = serial.Serial()
        self.socket.port     = dev
        self.socket.baudrate = rate
        self.socket.parity   = 'N'
        self.socket.rtscts   = False
        self.socket.xonxoff  = False
        self.socket.timeout  = 0.01
        self.version = sys.version_info[0]
        try:
            self.socket.open()
        except serial.SerialException:
            print('Could not open serial port {:s}'.format(self.socket.portstr))
            exit()

    def __del__(self):
        self.socket.close()

    def close(self):
        self.socket.close()

    def open(self):
        self.socket.open()

    def flush(self):
        self.socket.reset_input_buffer()
        self.socket.reset_output_buffer()

    def read(self):
        try:
            data = self.socket.read()
            return bytearray(data) if self.version == 2 else data
        except serial.SerialException:
            print('Serial port error')
            exit()

    def write(self, data):
        self.socket.write(data)


def get_all_serial_devices():
    return serial.tools.list_ports()



def _get_multiline_format(string):
    """
    Returns either `\r\n`, `\r`, `\n`, or None, if unable to detect
    """

    if "\r\n" in string:
        return r"\r\n"
    elif "\n" in string:
        return r"\n"
    elif "\r" in string:
        return r"\r"
    else:
        return None


def iterate_string_multiline(string: str, min_n_newline_symbols=1):
    import re

    multiline_format = _get_multiline_format(string)

    if multiline_format is None:
        # There are no multiline splits
        if len(string) > 0:
            yield string

        return

    # Match
    regex = '(' + multiline_format + ')' + "{%d,}" % min_n_newline_symbols
    regex_matcher = re.compile(regex, re.MULTILINE)
    text_body_position_begin = 0

    for m in re.finditer(regex, string):
        text_body_position_end, next_test_body_position_begin = m.span(0)
        chunk = string[text_body_position_begin:text_body_position_end]
        text_body_position_begin = next_test_body_position_begin

        if len(chunk):
            yield chunk

    chunk = string[text_body_position_begin:]

    if len(chunk) > 0:
        yield(chunk)




from setuptools import setup

def get_long_description():
    with open("README.md", 'r', encoding='utf-8') as f:
        return f.read()

setup(
    name="tired",
    packages=[
        "tired"
    ],
    include_package_data=True,
    license="MIT",
    description="Boilerplate I'm tired of writing over and over",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/damurashov/TIRED",
    author="Dmitry Murashov",
    setup_requires=["wheel"],
    install_requires=[
        "simple_term_menu",
        "datetime",
        "pyserial"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    version="0.3.3",
)




import subprocess
from shlex import split
import tired.logging


_CONTEXT = "tired.command"


def listout(l: list):
	command_string = '> ' + ' '.join(l)
	tired.logging.debug(_CONTEXT, f"Executing command `{command_string}`")


def get_output_piped(commands, verbose=False):
	"""
	:param commands: commands to execute. The commands are executed in a sequential manner, i.e. piped
	:return: console output, string
	"""
	commands = [split(c) for c in commands]
	procs = []

	listout(commands[0])
	p = subprocess.Popen(commands[0], stdout=subprocess.PIPE)

	for c in commands[1:]:
		tired.logging.debug(_CONTEXT, "  ↑ piped into ↓")
		listout(c)
		p = subprocess.Popen(c, stdout=subprocess.PIPE, stdin=p.stdout)

	p.wait()
	out = p.stdout.read()

	if verbose and len(out):
		print(out.decode("unicode_escape"))

	return out, p.returncode


def get_output(cmd):
	tired.logging.debug(_CONTEXT, "Executing command " + '>  `' + cmd + '`')
	ret = subprocess.run(split(cmd), stdout=subprocess.PIPE)

	if ret.returncode != 0:
		raise Exception("Command execution error")

	return ret.stdout.decode("unicode_escape").strip()


def get_output_with_code(cmd):
	tired.logging.debug(_CONTEXT, "Executing command " + '>  `' + cmd + '`')
	ret = subprocess.run(split(cmd), stdout=subprocess.PIPE)

	return ret.stdout.decode("unicode_escape").strip(), ret.returncode


def execute(cmd):
	tired.logging.debug(_CONTEXT, "Executing command " + '>  `' + cmd + '`')
	ret = subprocess.run(split(cmd))

	if ret.returncode != 0:
		raise Exception(f"Command execution error (returncode - {ret.returncode})")



import tired
import tired.datetime
import tired.meta


_LOG_SECTION_DELIMETER = "-"
NONE = 0
ERROR = 1
WARNING = 2
INFO = 3
DEBUG = 4
LOG_LEVEL_TO_STRING_MAPPING = {
    ERROR: "E",
    WARNING: "W",
    INFO: "I",
    DEBUG: "D"
}

def _log_impl(level, *args):
    context = tired.meta.get_stack_context_string(3)
    message = ' '.join(args)
    output = ' '.join([LOG_LEVEL_TO_STRING_MAPPING[level], _LOG_SECTION_DELIMETER,
        f"{tired.datetime.get_today_time_milliseconds_string()}", f"[{context}]", _LOG_SECTION_DELIMETER, message])
    print(output)


def debug(*args):
    _log_impl(DEBUG, *args)


def error(*args):
    _log_impl(ERROR, *args)


def info(*args):
    _log_impl(INFO, *args)


def warning(*args):
    _log_impl(WARNING, *args)
